import sys
from warnings import warn

from six import text_type as unicode
from six import PY2


def task_name(task):
    """
    Compute the task name for an Eliot task.

    If we can't find a ``message_type`` or an ``action_type`` field to use to
    derive the name, then return ``None``.
    """
    if task is None:
        raise ValueError('Cannot compute task name', task)
    level = u','.join(map(unicode, task[u'task_level']))
    message_type = task.get('message_type', None)
    if message_type is not None:
        status = u''
    elif message_type is None:
        message_type = task.get('action_type', None)
        if message_type is None:
            return None
        status = u'/' + task['action_status']
    return u'{message_type}@{level}{status}'.format(
        message_type=message_type,
        level=level,
        status=status)


class _TaskNode(object):
    """
    A node representing an Eliot task and its child tasks.

    :type task: ``dict``
    :ivar task: Eliot task.

    :type name: ``unicode``
    :ivar name: Node name; this will be derived from the task if it is not
        specified.

    :type _children: ``dict`` of ``_TaskNode``
    :ivar _children: Child nodes, see ``_TaskNode.children``
    """

    _DEFAULT_TASK_NAME = u'<UNNAMED TASK>'

    def __init__(self, task, name=None):
        if task is None:
            raise ValueError('Missing eliot task')
        self.task = task
        self._children = dict()
        if name is None:
            name = task_name(task) or self._DEFAULT_TASK_NAME
        self.name = name
        self.success = None

    def __repr__(self):
        """
        Human-readable representation of the node.
        """
        task_uuid = self.task[u'task_uuid']
        name = self.name
        if PY2:
            # XXX: This is probably wrong in a bunch of places.
            task_uuid = task_uuid.encode('utf-8')
            name = name.encode('utf-8')
        return '<{type} {task_uuid!r} {name!r} children={children}>'.format(
            type=type(self).__name__,
            task_uuid=task_uuid,
            name=name,
            children=len(self._children))

    def copy(self):
        """
        Make a shallow copy of this node.
        """
        return type(self)(self.task, self.name)

    def add_child(self, node):
        """
        Add a child node.

        :type node: ``_TaskNode``
        :param node: Child node to add to the tree, if the child has multiple
            levels it may be added as a grandchild.
        """
        def _add_child(parent, levels):
            levels = list(levels)
            level = levels.pop(0)
            children = parent._children
            if level in children:
                _add_child(children[level], levels)
            else:
                children[level] = node
                action_status = node.task.get('action_status')
                if action_status == u'succeeded':
                    node.success = parent.success = True
                elif action_status == u'failed':
                    node.success = parent.success = False
        _add_child(self, node.task['task_level'])

    def children(self):
        """
        Get a ``list`` of child ``_TaskNode``s ordered by task level.
        """
        return sorted(
            self._children.values(), key=lambda n: n.task[u'task_level'])


def missing_start_task(task_missing_parent):
    """
    Create a fake start task for an existing task that happens to be missing
    one.
    """
    return {
        u'action_type': u'<missing start task>',
        u'action_status': u'started',
        u'timestamp': task_missing_parent[u'timestamp'],
        u'task_uuid': task_missing_parent[u'task_uuid'],
        u'task_level': [1]}


class TaskMergeError(RuntimeError):
    """
    An exception occured while trying to merge a task into the tree.
    """
    def __init__(self, task, exc_info):
        self.task = task
        self.exc_info = exc_info
        RuntimeError.__init__(self)


class Tree(object):
    """
    Eliot task tree.

    :ivar _nodes: Internal tree storage, use ``Tree.nodes`` or
        ``Tree.matching_nodes`` to obtain the tree nodes.
    """
    def __init__(self):
        warn('Tree is deprecated, use eliottree.tasks_from_iterable instead',
             DeprecationWarning, 2)
        self._nodes = {}

    def nodes(self, uuids=None):
        """
        All top-level nodes in the tree.

        :type uuids: ``set`` of ``unicode``
        :param uuids: Set of task UUIDs to include, or ``None`` for no
            filtering.

        :rtype: ``iterable`` of 2-``tuple``s
        :return: Iterable of key and node pairs for top-level nodes, sorted by
            timestamp.
        """
        if uuids is not None:
            nodes = ((k, self._nodes[k]) for k in uuids)
        else:
            nodes = self._nodes.items()
        return sorted(nodes, key=lambda x: x[1].task[u'timestamp'])

    def merge_tasks(self, tasks, filter_funcs=None):
        """
        Merge tasks into the tree.

        :type tasks: ``iterable`` of ``dict``
        :param tasks: Iterable of task dicts.

        :type filter_funcs: ``iterable`` of 1-argument ``callable``s returning
            ``bool``
        :param filter_funcs: Iterable of predicate functions that given a task
            determine whether to keep it.

        :return: ``set`` of task UUIDs that match all of the filter functions,
            can be passed to ``Tree.matching_nodes``, or ``None`` if no filter
            functions were specified.
        """
        tasktree = self._nodes
        if filter_funcs is None:
            filter_funcs = []
        filter_funcs = list(filter_funcs)
        matches = dict((i, set()) for i, _ in enumerate(filter_funcs))

        def _merge_one(task, create_missing_tasks):
            key = task[u'task_uuid']
            node = tasktree.get(key)
            if node is None:
                if task[u'task_level'] != [1]:
                    if create_missing_tasks:
                        n = tasktree[key] = _TaskNode(
                            task=missing_start_task(task))
                        n.add_child(_TaskNode(task))
                    else:
                        return task
                else:
                    node = tasktree[key] = _TaskNode(task=task)
            else:
                node.add_child(_TaskNode(task))
            for i, fn in enumerate(filter_funcs):
                if fn(task):
                    matches[i].add(key)
            return None

        def _merge(tasks, create_missing_tasks=False):
            pending = []
            for task in tasks:
                try:
                    result = _merge_one(task, create_missing_tasks)
                    if result is not None:
                        pending.append(result)
                except Exception:
                    raise TaskMergeError(task, sys.exc_info())
            return pending

        pending = _merge(tasks)
        if pending:
            pending = _merge(pending, True)
            if pending:
                raise RuntimeError('Some tasks have no start parent', pending)
        if not matches:
            return None
        return set.intersection(*matches.values())


__all__ = ['Tree']



import colored as _colored


attr_codes = {
    'bold': 1,
    'dim': 2,
    'underlined': 4,
    'blink': 5,
    'reverse': 7,
    'hidden': 8,
    'reset': 0,
    'res_bold': 21,
    'res_dim': 22,
    'res_underlined': 24,
    'res_blink': 25,
    'res_reverse': 27,
    'res_hidden': 28,
}


def colored(text, fg=None, bg=None, attrs=None):
    """
    Wrap text in terminal color codes.

    Intended to mimic the API of `termcolor`.
    """
    if attrs is None:
        attrs = []
    return u'{}{}{}{}{}'.format(
        _colored.fg(fg) if fg is not None else u'',
        _colored.bg(bg) if bg is not None else u'',
        text,
        u''.join(_colored.attr(attr_codes.get(attr, attr)) for attr in attrs),
        _colored.attr(0))


def color_factory(colored):
    """
    Factory for making text color-wrappers.
    """
    def _color(fg, bg=None, attrs=[]):
        def __color(text):
            return colored(text, fg, bg, attrs=attrs)
        return __color
    return _color



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Jonathan M. Lange <jml@mumak.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Library for formatting trees."""

import itertools


RIGHT_DOUBLE_ARROW = u'\N{RIGHTWARDS DOUBLE ARROW}'
HOURGLASS = u'\N{WHITE HOURGLASS}'


class Options(object):
    def __init__(self,
                 FORK=u'\u251c',
                 LAST=u'\u2514',
                 VERTICAL=u'\u2502',
                 HORIZONTAL=u'\u2500',
                 NEWLINE=u'\u23ce',
                 ARROW=RIGHT_DOUBLE_ARROW,
                 HOURGLASS=HOURGLASS):
        self.FORK = FORK
        self.LAST = LAST
        self.VERTICAL = VERTICAL
        self.HORIZONTAL = HORIZONTAL
        self.NEWLINE = NEWLINE
        self.ARROW = ARROW
        self.HOURGLASS = HOURGLASS

    def color(self, node, depth):
        return lambda text, *a, **kw: text

    def vertical(self):
        return u''.join([self.VERTICAL, u'   '])

    def fork(self):
        return u''.join([self.FORK, self.HORIZONTAL, self.HORIZONTAL, u' '])

    def last(self):
        return u''.join([self.LAST, self.HORIZONTAL, self.HORIZONTAL, u' '])



ASCII_OPTIONS = Options(FORK=u'|',
                        LAST=u'+',
                        VERTICAL=u'|',
                        HORIZONTAL=u'-',
                        NEWLINE=u'\n',
                        ARROW=u'=>',
                        HOURGLASS='|Y|')


def _format_newlines(prefix, formatted_node, options):
    """
    Convert newlines into U+23EC characters, followed by an actual newline and
    then a tree prefix so as to position the remaining text under the previous
    line.
    """
    replacement = u''.join([
        options.NEWLINE,
        u'\n',
        prefix])
    return formatted_node.replace(u'\n', replacement)


def _format_tree(node, format_node, get_children, options, prefix=u'', depth=0):
    children = list(get_children(node))
    color = options.color(node, depth)
    #options.set_depth(depth)
    next_prefix = prefix + color(options.vertical())
    for child in children[:-1]:
        yield u''.join([prefix,
                        color(options.fork()),
                        _format_newlines(next_prefix,
                                         format_node(child),
                                         options)])
        for result in _format_tree(child,
                                   format_node,
                                   get_children,
                                   options,
                                   next_prefix,
                                   depth=depth + 1):
            yield result
    if children:
        last_prefix = u''.join([prefix, u'    '])
        yield u''.join([prefix,
                        color(options.last()),
                        _format_newlines(last_prefix,
                                         format_node(children[-1]),
                                         options)])
        for result in _format_tree(children[-1],
                                   format_node,
                                   get_children,
                                   options,
                                   last_prefix,
                                   depth=depth + 1):
            yield result


def format_tree(node, format_node, get_children, options=None):
    lines = itertools.chain(
        [format_node(node)],
        _format_tree(node, format_node, get_children, options or Options()),
        [u''],
    )
    return u'\n'.join(lines)


def format_ascii_tree(tree, format_node, get_children):
    """ Formats the tree using only ascii characters """
    return format_tree(tree,
                       format_node,
                       get_children,
                       ASCII_OPTIONS)


def print_tree(*args, **kwargs):
    print(format_tree(*args, **kwargs))



import time
from testtools import TestCase
from testtools.matchers import Is

from eliottree import format
from eliottree.test.matchers import ExactlyEquals


class BinaryTests(TestCase):
    """
    Tests for `eliottree.format.binary`.
    """
    def test_not_binary(self):
        """
        Not binary values are ignored.
        """
        self.assertThat(
            format.binary('utf-8')(u'hello'),
            Is(None))

    def test_encoding(self):
        """
        Binary values are decoded with the given encoding.
        """
        self.assertThat(
            format.binary('utf-8')(u'\N{SNOWMAN}'.encode('utf-8')),
            ExactlyEquals(u'\u2603'))

    def test_replace(self):
        """
        Replace decoding errors with the Unicode replacement character.
        """
        self.assertThat(
            format.binary('utf-32')(u'\N{SNOWMAN}'.encode('utf-8')),
            ExactlyEquals(u'\ufffd'))


class TextTests(TestCase):
    """
    Tests for `eliottree.format.text`.
    """
    def test_not_text(self):
        """
        Not text values are ignored.
        """
        self.assertThat(
            format.text()(b'hello'),
            Is(None))

    def test_text(self):
        """
        Text values are returned as is.
        """
        self.assertThat(
            format.text()(u'\N{SNOWMAN}'),
            ExactlyEquals(u'\N{SNOWMAN}'))


class FieldsTests(TestCase):
    """
    Tests for `eliottree.format.fields`.
    """
    def test_missing_mapping(self):
        """
        Values for unknown field names are ignored.
        """
        self.assertThat(
            format.fields({})(b'hello', u'a'),
            Is(None))

    def test_mapping(self):
        """
        Values for known field names are passed through their processor.
        """
        fields = {
            u'a': format.binary('utf-8'),
        }
        self.assertThat(
            format.fields(fields)(u'\N{SNOWMAN}'.encode('utf-8'), u'a'),
            ExactlyEquals(u'\N{SNOWMAN}'))


class TimestampTests(TestCase):
    """
    Tests for `eliottree.format.timestamp`.
    """
    def test_text(self):
        """
        Timestamps as text are converted to ISO 8601.
        """
        self.assertThat(
            format.timestamp()(u'1433631432'),
            ExactlyEquals(u'2015-06-06 22:57:12Z'))

    def test_float(self):
        """
        Timestamps as floats are converted to ISO 8601.
        """
        self.assertThat(
            format.timestamp()(1433631432.0),
            ExactlyEquals(u'2015-06-06 22:57:12Z'))

    def test_local(self):
        """
        Timestamps can be converted to local time.
        """
        timestamp = 1433631432.0
        utc = format.timestamp(utc_timestamps=True)(timestamp)
        local = format.timestamp(utc_timestamps=False)(timestamp + time.timezone)
        # Strip the "Z" off the end.
        self.assertThat(utc[:-1], ExactlyEquals(local))


class AnythingTests(TestCase):
    """
    Tests for `eliottree.format.anything`.
    """
    def test_anything(self):
        """
        Convert values to Unicode via `repr`.
        """
        class _Foo(object):
            def __repr__(self):
                return 'hello'
        self.assertThat(
            format.anything('utf-8')(_Foo()),
            ExactlyEquals(u'hello'))
        self.assertThat(
            format.anything('utf-8')(42),
            ExactlyEquals(u'42'))


class TruncateTests(TestCase):
    """
    Tests for `eliottree.format.truncate_value`.
    """
    def test_under(self):
        """
        No truncation occurs if the length of the value is below the limit.
        """
        self.assertThat(
            format.truncate_value(10, u'abcdefghijklmnopqrstuvwxyz'[:5]),
            ExactlyEquals(u'abcde'))

    def test_exact(self):
        """
        No truncation occurs if the length of the value is exactly at the
        limit.
        """
        self.assertThat(
            format.truncate_value(10, u'abcdefghijklmnopqrstuvwxyz'[:10]),
            ExactlyEquals(u'abcdefghij'))

    def test_over(self):
        """
        Truncate values whose length is above the limit.
        """
        self.assertThat(
            format.truncate_value(10, u'abcdefghijklmnopqrstuvwxyz'),
            ExactlyEquals(u'abcdefghij\u2026'))

    def test_multiple_lines(self):
        """
        Truncate values that have more than a single line of text by only
        showing the first line.
        """
        self.assertThat(
            format.truncate_value(10, u'abc\ndef'),
            ExactlyEquals(u'abc\u2026'))



from datetime import datetime

from six import binary_type, text_type, unichr
from toolz import merge


_control_equivalents = dict((i, unichr(0x2400 + i)) for i in range(0x20))
_control_equivalents[0x7f] = u'\u2421'


def escape_control_characters(s, overrides={}):
    """
    Replace terminal control characters with their Unicode control character
    equivalent.
    """
    return text_type(s).translate(merge(_control_equivalents, overrides))


def some(*fs):
    """
    Create a function that returns the first non-``None`` result of applying
    the arguments to each ``fs``.
    """
    def _some(*a, **kw):
        for f in fs:
            result = f(*a, **kw)
            if result is not None:
                return result
        return None
    return _some


def binary(encoding):
    """
    Create a formatter for ``binary_type`` values.

    :param str encoding: Encoding to assume for values.
    """
    def _format_bytes_value(value, field_name=None):
        if isinstance(value, binary_type):
            return value.decode(encoding, 'replace')
    return _format_bytes_value


def text():
    """
    Create a formatter for ``text_type`` values.
    """
    def _format_text_value(value, field_name=None):
        if isinstance(value, text_type):
            return value
    return _format_text_value


def fields(format_mapping):
    """
    Create a formatter that performs specific formatting based on field names.

    :type format_mapping: ``Dict[text_type, Callable[[Any, text_type], Any]]``
    """
    def _format_field_value(value, field_name=None):
        f = format_mapping.get(field_name, None)
        if f is None:
            return None
        return f(value, field_name)
    return _format_field_value


def timestamp(include_microsecond=True, utc_timestamps=True):
    """
    Create a formatter for POSIX timestamp values.
    """
    def _format_timestamp_value(value, field_name=None):
        from_timestamp = (
            datetime.utcfromtimestamp
            if utc_timestamps else datetime.fromtimestamp)
        result = from_timestamp(float(value))
        if not include_microsecond:
            result = result.replace(microsecond=0)
        result = result.isoformat(' ')
        if isinstance(result, binary_type):
            result = result.decode('ascii')
        return result + (u'Z' if utc_timestamps else u'')
    return _format_timestamp_value


def duration():
    """
    Create a formatter for duration values specified as seconds.
    """
    def _format_duration(value, field_name=None):
        return u'{:.3f}s'.format(value)
    return _format_duration


def anything(encoding):
    """
    Create a formatter for any value using `repr`.

    :param str encoding: Encoding to assume for a `binary_type` result.
    """
    def _format_other_value(value, field_name=None):
        result = repr(value)
        if isinstance(result, binary_type):
            result = result.decode(encoding, 'replace')
        return result
    return _format_other_value


def truncate_value(limit, value):
    """
    Truncate ``value`` to a maximum of ``limit`` characters.
    """
    values = value.split(u'\n')
    value = values[0]
    if len(value) > limit or len(values) > 1:
        return u'{}\u2026'.format(value[:limit])
    return value


__all__ = [
    'escape_control_characters', 'some', 'binary', 'text', 'fields',
    'timestamp', 'anything', 'truncate_value']



# Copyright (c) 2015 Jonathan M. Lange <jml@mumak.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Nicely format trees."""

from ._text import (
    format_ascii_tree,
    format_tree,
    print_tree,
    Options,
    ASCII_OPTIONS,
)

__all__ = [
    'format_ascii_tree', 'format_tree', 'print_tree', 'Options', 'ASCII_OPTIONS']



from eliottree._color import color_factory


NO_COLOR = (None,)
DEFAULT_THEME = {
    # Task root.
    'root': NO_COLOR,
    # Action / message node.
    'parent': NO_COLOR,
    # Action / message task level.
    'task_level': NO_COLOR,
    # Action success status.
    'status_success': NO_COLOR,
    # Action failure status.
    'status_failure': NO_COLOR,
    # Task timestamp.
    'timestamp': NO_COLOR,
    # Action / message property key.
    'prop_key': NO_COLOR,
    # Action / message property value.
    'prop_value': NO_COLOR,
    # Task duration.
    'duration': NO_COLOR,
    # Tree color for failed tasks.
    'tree_failed': NO_COLOR,
    # Cycled tree colors.
    'tree_color0': NO_COLOR,
    'tree_color1': NO_COLOR,
    'tree_color2': NO_COLOR,
    # Processing error.
    'error': NO_COLOR,
}


class Theme(object):
    """
    Theme base class.
    """
    __slots__ = DEFAULT_THEME.keys()

    def __init__(self, color, **theme):
        super(Theme, self).__init__()
        self.color = color
        _theme = dict(DEFAULT_THEME)
        _theme.update(theme)
        for k, v in _theme.items():
            if not isinstance(v, (tuple, list)):
                raise TypeError(
                    'Theme color must be a tuple or list of values', k, v)
            setattr(self, k, color(*v))


class DarkBackgroundTheme(Theme):
    """
    Color theme for dark backgrounds.
    """
    def __init__(self, colored):
        super(DarkBackgroundTheme, self).__init__(
            color=color_factory(colored),
            root=('white', None, ['bold']),
            parent=('magenta',),
            status_success=('green',),
            status_failure=('red',),
            prop_key=('blue',),
            error=('red', None, ['bold']),
            timestamp=('white', None, ['dim']),
            duration=('blue', None, ['dim']),
            tree_failed=('red',),
            tree_color0=('white', None, ['dim']),
            tree_color1=('blue', None, ['dim']),
            tree_color2=('magenta', None, ['dim']),
        )


class LightBackgroundTheme(Theme):
    """
    Color theme for light backgrounds.
    """
    def __init__(self, colored):
        super(LightBackgroundTheme, self).__init__(
            color=color_factory(colored),
            root=('dark_gray', None, ['bold']),
            parent=('magenta',),
            status_success=('green',),
            status_failure=('red',),
            prop_key=('blue',),
            error=('red', None, ['bold']),
            timestamp=('dark_gray',),
            duration=('blue', None, ['dim']),
            tree_failed=('red',),
            tree_color0=('dark_gray', None, ['dim']),
            tree_color1=('blue', None, ['dim']),
            tree_color2=('magenta', None, ['dim']),
        )


def _no_color(text, *a, **kw):
    """
    Colorizer that does not colorize.
    """
    return text


def get_theme(dark_background, colored=None):
    """
    Create an appropriate theme.
    """
    if colored is None:
        colored = _no_color
    return DarkBackgroundTheme(colored) if dark_background else LightBackgroundTheme(colored)


def apply_theme_overrides(theme, overrides):
    """
    Apply overrides to a theme.
    """
    if overrides is None:
        return theme

    for key, args in overrides.items():
        setattr(theme, key, theme.color(*args))
    return theme



import json

from six import text_type


def dump_json_bytes(obj, dumps=json.dumps):
    """
    Serialize ``obj`` to JSON formatted UTF-8 encoded ``bytes``.
    """
    result = dumps(obj)
    if isinstance(result, text_type):
        return result.encode('utf-8')
    return result



from testtools import ExpectedException, TestCase
from testtools.matchers import AfterPreprocessing as After
from testtools.matchers import (
    Equals, Is, MatchesAll, MatchesListwise, MatchesPredicate,
    MatchesStructure)

from eliottree._util import format_namespace, is_namespace, namespaced


class NamespaceTests(TestCase):
    """
    Tests for `namespaced`, `format_namespace` and `is_namespace`.
    """
    def test_namespaced(self):
        """
        `namespaced` creates a function that when called produces a namespaced
        name.
        """
        self.assertThat(
            namespaced(u'foo'),
            MatchesAll(
                MatchesPredicate(callable, '%s is not callable'),
                After(
                    lambda f: f(u'bar'),
                    MatchesAll(
                        MatchesListwise([
                            Equals(u'foo'),
                            Equals(u'bar')]),
                        MatchesStructure(
                            prefix=Equals(u'foo'),
                            name=Equals(u'bar'))))))

    def test_format_not_namespace(self):
        """
        `format_namespace` raises `TypeError` if its argument is not a
        namespaced name.
        """
        with ExpectedException(TypeError):
            format_namespace(42)

    def test_format_namespace(self):
        """
        `format_namespace` creates a text representation of a namespaced name.
        """
        self.assertThat(
            format_namespace(namespaced(u'foo')(u'bar')),
            Equals(u'foo/bar'))

    def test_is_namespace(self):
        """
        `is_namespace` returns ``True`` only for namespaced names.
        """
        self.assertThat(
            is_namespace(42),
            Is(False))
        self.assertThat(
            is_namespace((u'foo', u'bar')),
            Is(False))
        self.assertThat(
            is_namespace(namespaced(u'foo')),
            Is(False))
        self.assertThat(
            is_namespace(namespaced(u'foo')(u'bar')),
            Is(True))



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015 Jonathan M. Lange <jml@mumak.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import doctest
from operator import itemgetter
from textwrap import dedent

from testtools import TestCase
from testtools.matchers import DocTestMatches

from .._text import (
    format_tree, format_ascii_tree,
)


class TestFormatTree(TestCase):

    def format_tree(self, tree):
        return format_tree(tree, itemgetter(0), itemgetter(1))

    def format_ascii_tree(self, tree):
        return format_ascii_tree(tree, itemgetter(0), itemgetter(1))

    def test_single_node_tree(self):
        tree = ('foo', [])
        output = self.format_tree(tree)
        self.assertEqual(dedent(u'''\
        foo
        '''), output)

    def test_single_level_tree(self):
        tree = (
            'foo', [
                ('bar', []),
                ('baz', []),
                ('qux', []),
            ],
        )
        output = self.format_tree(tree)
        self.assertEqual(dedent(u'''\
        foo
        ├── bar
        ├── baz
        └── qux
        '''), output)

    def test_multi_level_tree(self):
        tree = (
            'foo', [
                ('bar', [
                    ('a', []),
                    ('b', []),
                ]),
                ('baz', []),
                ('qux', []),
            ],
        )
        output = self.format_tree(tree)
        self.assertEqual(dedent(u'''\
        foo
        ├── bar
        │   ├── a
        │   └── b
        ├── baz
        └── qux
        '''), output)

    def test_multi_level_on_last_node_tree(self):
        tree = (
            'foo', [
                ('bar', []),
                ('baz', []),
                ('qux', [
                    ('a', []),
                    ('b', []),
                ]),
            ],
        )
        output = self.format_tree(tree)
        self.assertEqual(dedent(u'''\
        foo
        ├── bar
        ├── baz
        └── qux
            ├── a
            └── b
        '''), output)

    def test_acceptance(self):
        output = self.format_tree(ACCEPTANCE_INPUT)
        self.assertThat(
            output,
            DocTestMatches(
                ACCEPTANCE_OUTPUT,
                doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE | doctest.REPORT_NDIFF))

    def test_newlines(self):
        tree = (
            'foo', [
                ('bar\nfrob', [
                    ('a', []),
                    ('b\nc\nd', []),
                ]),
                ('baz', []),
                ('qux\nfrab', []),
            ],
        )
        output = self.format_tree(tree)
        self.assertEqual(dedent(u'''\
        foo
        ├── bar⏎
        │   frob
        │   ├── a
        │   └── b⏎
        │       c⏎
        │       d
        ├── baz
        └── qux⏎
            frab
        '''), output)

    def test_ascii__tree(self):
        tree = (
            'foo', [
                ('bar', [
                    ('a', []),
                    ('b', []),
                ]),
                ('baz', []),
                ('qux', []),
            ],
        )
        output = self.format_ascii_tree(tree)
        self.assertEqual(dedent(u'''\
        foo
        |-- bar
        |   |-- a
        |   +-- b
        |-- baz
        +-- qux
        '''), output)


def d(name, files):
    return (name, files)


def f(name):
    return (name, [])


ACCEPTANCE_INPUT = \
    d(u'.', [
        f(u'cabal.sandbox.config'),
        f(u'config.yaml'),
        d(u'dist', [
            d(u'build', [
                d(u'autogen', [
                    f(u'cabal_macros.h'),
                    f(u'Paths_hodor.hs'),
                ]),
                d(u'hodor', [
                    f(u'hodor'),
                    d(u'hodor-tmp', [
                        d(u'Hodor', map(f, [
                            u'Actions.hi',
                            u'Actions.o',
                            u'CommandLine.hi',
                            u'CommandLine.o',
                            u'Commands.hi',
                            u'Commands.o',
                            u'Config.hi',
                            u'Config.o',
                            u'File.hi',
                            u'File.o',
                            u'Format.hi',
                            u'Format.o',
                            u'Functional.hi',
                            u'Functional.o',
                            u'Parser.hi',
                            u'Parser.o',
                            u'Types.hi',
                            u'Types.o',
                        ])),
                        f(u'Hodor.hi'),
                        f(u'Hodor.o'),
                        f(u'Main.hi'),
                        f(u'Main.o'),
                    ]),
                ]),
            ]),
            f(u'package.conf.inplace'),
            f(u'setup-config'),
        ]),
        d(u'Hodor', map(f, [
            u'Actions.hs',
            u'CommandLine.hs',
            u'Commands.hs',
            u'Config.hs',
            u'File.hs',
            u'Format.hs',
            u'Functional.hs',
            u'Parser.hs',
            u'Reports.hs',
            u'Types.hs',
        ])),
        f(u'hodor.cabal'),
        f(u'Hodor.hs'),
        f(u'LICENSE'),
        f(u'Main.hs'),
        f(u'notes.md'),
        f(u'Setup.hs'),
        d(u'Tests', map(f, [
            u'FormatSpec.hs',
            u'Generators.hs',
            u'ParserSpec.hs',
            u'TypesSpec.hs',
        ])),
        f(u'Tests.hs'),
        f(u'todo.txt'),
    ])


ACCEPTANCE_OUTPUT = u'''\
.
├── cabal.sandbox.config
├── config.yaml
├── dist
│   ├── build
│   │   ├── autogen
│   │   │   ├── cabal_macros.h
│   │   │   └── Paths_hodor.hs
│   │   └── hodor
│   │       ├── hodor
│   │       └── hodor-tmp
│   │           ├── Hodor
│   │           │   ├── Actions.hi
│   │           │   ├── Actions.o
│   │           │   ├── CommandLine.hi
│   │           │   ├── CommandLine.o
│   │           │   ├── Commands.hi
│   │           │   ├── Commands.o
│   │           │   ├── Config.hi
│   │           │   ├── Config.o
│   │           │   ├── File.hi
│   │           │   ├── File.o
│   │           │   ├── Format.hi
│   │           │   ├── Format.o
│   │           │   ├── Functional.hi
│   │           │   ├── Functional.o
│   │           │   ├── Parser.hi
│   │           │   ├── Parser.o
│   │           │   ├── Types.hi
│   │           │   └── Types.o
│   │           ├── Hodor.hi
│   │           ├── Hodor.o
│   │           ├── Main.hi
│   │           └── Main.o
│   ├── package.conf.inplace
│   └── setup-config
├── Hodor
│   ├── Actions.hs
│   ├── CommandLine.hs
│   ├── Commands.hs
│   ├── Config.hs
│   ├── File.hs
│   ├── Format.hs
│   ├── Functional.hs
│   ├── Parser.hs
│   ├── Reports.hs
│   └── Types.hs
├── hodor.cabal
├── Hodor.hs
├── LICENSE
├── Main.hs
├── notes.md
├── Setup.hs
├── Tests
│   ├── FormatSpec.hs
│   ├── Generators.hs
│   ├── ParserSpec.hs
│   └── TypesSpec.hs
├── Tests.hs
└── todo.txt
'''



from collections import namedtuple


namespace = namedtuple('namespace', ['prefix', 'name'])


def namespaced(prefix):
    """
    Create a function that creates new names in the ``prefix`` namespace.

    :rtype: Callable[[unicode], `namespace`]
    """
    return lambda name: namespace(prefix, name)


def format_namespace(ns):
    """
    Format a `namespace`.

    :rtype: unicode
    """
    if not is_namespace(ns):
        raise TypeError('Expected namespace', ns)
    return u'{}/{}'.format(ns.prefix, ns.name)


def is_namespace(x):
    """
    Is this a `namespace` instance?

    :rtype: bool
    """
    return isinstance(x, namespace)


eliot_ns = namespaced(u'eliot')



"""
Tests for the command-line itself.
"""
import os
import six
import tempfile
from collections import namedtuple
from subprocess import PIPE, CalledProcessError, Popen
from unittest import TestCase

from eliottree._compat import dump_json_bytes
from eliottree.test.tasks import message_task, missing_uuid_task


rendered_message_task = (
    u'cdeb220d-7605-4d5f-8341-1a170222e308\n'
    u'\u2514\u2500\u2500 twisted:log/1 2015-03-03 04:25:00Z\n'
    u'    \u251c\u2500\u2500 error: False\n'
    u'    \u2514\u2500\u2500 message: Main loop terminated.\n\n'
).replace('\n', os.linesep).encode('utf-8')


def bytes_hex(a):
    """
    Encode `bytes` into hex on Python 2 or 3.
    """
    if hasattr(a, 'hex'):
        return a.hex()
    return a.encode('hex')


class NamedTemporaryFile(object):
    """
    Similar to `tempfile.NamedTemporaryFile` except less buggy and with a
    Python context manager interface.
    """
    def __init__(self):
        # See https://stackoverflow.com/a/58955530/81065
        self.name = os.path.join(tempfile.gettempdir(), bytes_hex(os.urandom(24)))
        self._fd = open(self.name, 'wb+')
        self.write = self._fd.write
        self.flush = self._fd.flush
        self.seek = self._fd.seek

    def close(self):
        self._fd.close()
        if os.path.exists(self.name):
            os.unlink(self.name)

    def __enter__(self):
        return self._fd

    def __exit__(self, type, value, traceback):
        self.flush()
        self.close()


def check_output(args, stdin=None):
    """
    Similar to `subprocess.check_output` but include separated stdout and
    stderr in the `CalledProcessError` exception.
    """
    kwargs = {}
    pipes = Popen(
        args,
        stdin=PIPE if stdin is not None else None,
        stdout=PIPE,
        stderr=PIPE,
        **kwargs)
    stdout, stderr = pipes.communicate(
        six.ensure_binary(stdin) if stdin is not None else None)
    if pipes.returncode != 0:
        output = namedtuple('Output', ['stdout', 'stderr'])(
            six.ensure_binary(stdout),
            six.ensure_binary(stderr))
        raise CalledProcessError(pipes.returncode, args, output)
    return six.ensure_binary(stdout)


class EndToEndTests(TestCase):
    """
    Tests that actually run the command-line tool.
    """
    def test_stdin(self):
        """
        ``eliot-tree`` can read and render JSON messages from stdin when no
        arguments are given.
        """
        with NamedTemporaryFile() as f:
            f.write(dump_json_bytes(message_task))
            f.flush()
            f.seek(0, 0)
            self.assertEqual(check_output(["eliot-tree"], stdin=f.read()),
                             rendered_message_task)

    def test_file(self):
        """
        ``eliot-tree`` can read and render JSON messages from a file on the
        command line.
        """
        with NamedTemporaryFile() as f:
            f.write(dump_json_bytes(message_task))
            f.flush()
            self.assertEqual(check_output(["eliot-tree", f.name]),
                             rendered_message_task)

    def test_json_parse_error(self):
        """
        ``eliot-tree`` displays an error containing the file name, line number
        and offending line in the event that JSON parsing fails.
        """
        with NamedTemporaryFile() as f:
            f.write(dump_json_bytes(message_task) + b'\n')
            f.write(b'totally not valid JSON {')
            f.flush()
            with self.assertRaises(CalledProcessError) as m:
                check_output(['eliot-tree', '--color=never', f.name])
            lines = m.exception.output.stderr.splitlines()
            first_line = lines[0].decode('utf-8')
            second_line = lines[1].decode('utf-8')
            self.assertIn('JSON parse error', first_line)
            self.assertIn(f.name, first_line)
            self.assertIn('line 2', first_line)
            self.assertEqual('totally not valid JSON {', second_line)

    def test_eliot_parse_error(self):
        """
        ``eliot-tree`` displays an error containing the original file name,
        line number and offending task in the event that parsing the message
        dict fails.
        """
        with NamedTemporaryFile() as f:
            f.write(dump_json_bytes(missing_uuid_task) + b'\n')
            f.flush()
            with self.assertRaises(CalledProcessError) as m:
                check_output(['eliot-tree', f.name])
            lines = m.exception.output.stderr.splitlines()
            first_line = lines[0].decode('utf-8')
            self.assertIn('Eliot message parse error', first_line)
            self.assertIn(f.name, first_line)
            self.assertIn('line 1', first_line)



from testtools.matchers import Equals, IsInstance, MatchesAll


def ExactlyEquals(value):
    """
    Like `Equals` but also requires that the types match.
    """
    return MatchesAll(
        IsInstance(type(value)),
        Equals(value))



unnamed_message = {
    u"task_uuid": u"cdeb220d-7605-4d5f-8341-1a170222e308",
    u"error": False,
    u"timestamp": 1425356700,
    u"message": u"Main loop terminated.",
    u"task_level": [1]}

message_task = {
    u"task_uuid": u"cdeb220d-7605-4d5f-8341-1a170222e308",
    u"error": False,
    u"timestamp": 1425356700,
    u"message": u"Main loop terminated.",
    u"message_type": u"twisted:log",
    u"task_level": [1]}

action_task = {
    u"timestamp": 1425356800,
    u"action_status": u"started",
    u"task_uuid": u"f3a32bb3-ea6b-457c-aa99-08a3d0491ab4",
    u"action_type": u"app:action",
    u"task_level": [1]}

nested_action_task = {
    u"timestamp": 1425356900,
    u"action_status": u"started",
    u"task_uuid": u"f3a32bb3-ea6b-457c-aa99-08a3d0491ab4",
    u"action_type": u"app:action:nest",
    u"task_level": [1, 1]}

action_task_end = {
    u"timestamp": 1425356802,
    u"action_status": u"succeeded",
    u"task_uuid": u"f3a32bb3-ea6b-457c-aa99-08a3d0491ab4",
    u"action_type": u"app:action",
    u"task_level": [2]}

action_task_end_failed = {
    u"timestamp": 1425356804,
    u"action_status": u"failed",
    u"task_uuid": u"f3a32bb3-ea6b-457c-aa99-08a3d0491ab4",
    u"action_type": u"app:action",
    u"task_level": [2]}

dict_action_task = {
    u"timestamp": 1425356800,
    u"action_status": u"started",
    u"task_uuid": u"f3a32bb3-ea6b-457c-aa99-08a3d0491ab4",
    u"action_type": u"app:action",
    u"task_level": [1],
    u"some_data": {u"a": 42}}

list_action_task = {
    u"timestamp": 1425356800,
    u"action_status": u"started",
    u"task_uuid": u"f3a32bb3-ea6b-457c-aa99-08a3d0491ab4",
    u"action_type": u"app:action",
    u"task_level": [1],
    u"some_data": [u"a", u"b"]}

multiline_action_task = {
    u"timestamp": 1425356800,
    u"action_status": u"started",
    u"task_uuid": u"f3a32bb3-ea6b-457c-aa99-08a3d0491ab4",
    u"action_type": u"app:action",
    u"task_level": [1],
    u"message": u"this is a\nmany line message"}

janky_action_task = {
    u"timestamp": '1425356800\x1b(0',
    u"action_status": u"started",
    u"task_uuid": u"f3a32bb3-ea6b-457c-\x1b(0aa99-08a3d0491ab4",
    u"action_type": u"A\x1b(0",
    u"task_level": [1],
    u"mes\nsage": u"hello\x1b(0world",
    u"\x1b(0": {u"\x1b(0": "nope"}}

janky_message_task = {
    u"task_uuid": u"cdeb220d-7605-4d5f-\x1b(08341-1a170222e308",
    u"er\x1bror": False,
    u"timestamp": 1425356700,
    u"mes\nsage": u"Main loop\x1b(0terminated.",
    u"message_type": u"M\x1b(0",
    u"task_level": [1]}

missing_uuid_task = {
    u"error": False,
    u"timestamp": 1425356700,
    u"message": u"Main loop terminated.",
    u"message_type": u"twisted:log",
    u"action_type": u"nope",
    u"task_level": [1]}



import os
import codecs
import versioneer
from setuptools import setup, find_packages

HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(HERE, *parts), 'rb', 'utf-8') as f:
        return f.read()


setup(
    name='eliot-tree',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Render Eliot logs as an ASCII tree',
    license='MIT',
    url='https://github.com/jonathanj/eliottree',
    author='Jonathan Jacobs',
    author_email='jonathan@jsphere.com',
    maintainer='Jonathan Jacobs',
    maintainer_email='jonathan@jsphere.com',
    include_package_data=True,
    long_description=read('README.rst'),
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    entry_points={
        # These are the command-line programs we want setuptools to install.
        'console_scripts': [
            'eliot-tree = eliottree._cli:main',
        ],
    },
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: System :: Logging',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    install_requires=[
        'six>=1.13.0',
        'jmespath>=0.7.1',
        'iso8601>=0.1.10',
        'colored>=1.4.2',
        'toolz>=0.8.2',
        'eliot>=1.6.0',
        'win-unicode-console>=0.5;platform_system=="Windows"',
    ],
    extras_require={
        'test': ['testtools>=1.8.0'],
    },
)



import time
from eliot.parse import WrittenMessage
from six import StringIO, text_type
from testtools import ExpectedException, TestCase
from testtools.matchers import AfterPreprocessing as After
from testtools.matchers import (
    Contains, Equals, HasLength, MatchesAll, MatchesListwise, StartsWith)

from eliottree import (
    render_tasks, tasks_from_iterable)
from eliottree._color import colored
from eliottree._render import (
    _default_value_formatter, format_node,
    get_children, message_fields, message_name)
from eliottree.tree_format._text import (
    HOURGLASS, RIGHT_DOUBLE_ARROW, Options)
from eliottree._theme import get_theme
from eliottree._util import eliot_ns
from eliottree.test.matchers import ExactlyEquals
from eliottree.test.tasks import (
    action_task, action_task_end, action_task_end_failed, dict_action_task,
    janky_action_task, janky_message_task, list_action_task, message_task,
    multiline_action_task, nested_action_task)


class DefaultValueFormatterTests(TestCase):
    """
    Tests for ``eliottree.render._default_value_formatter``.
    """
    def test_unicode(self):
        """
        Pass ``unicode`` values straight through.
        """
        format_value = _default_value_formatter(
            human_readable=False, field_limit=0)
        self.assertThat(
            format_value(u'\N{SNOWMAN}'),
            ExactlyEquals(u'\N{SNOWMAN}'))

    def test_unicode_control_characters(self):
        """
        Translate control characters to their Unicode "control picture"
        equivalent, instead of destroying a terminal.
        """
        format_value = _default_value_formatter(
            human_readable=False, field_limit=0)
        self.assertThat(
            format_value(u'hello\001world'),
            ExactlyEquals(u'hello\u2401world'))

    def test_bytes(self):
        """
        Decode ``bytes`` values as the specified encoding.
        """
        format_value = _default_value_formatter(
            human_readable=False, field_limit=0, encoding='utf-8')
        self.assertThat(
            format_value(b'foo'),
            ExactlyEquals(u'foo'))
        self.assertThat(
            format_value(b'\xe2\x98\x83'),
            ExactlyEquals(u'\N{SNOWMAN}'))

    def test_anything(self):
        """
        Pass unknown values to ``repr``.
        """
        class _Thing(object):
            def __repr__(self):
                return 'Hello'
        format_value = _default_value_formatter(
            human_readable=False, field_limit=0)
        self.assertThat(
            format_value(_Thing()),
            ExactlyEquals(u'Hello'))

    def test_timestamp_field(self):
        """
        Format Eliot ``timestamp`` fields as human-readable if the feature was
        requested.
        """
        format_value = _default_value_formatter(
            human_readable=True, field_limit=0)
        # datetime(2015, 6, 6, 22, 57, 12)
        now = 1433631432
        self.assertThat(
            format_value(now, eliot_ns(u'timestamp')),
            ExactlyEquals(u'2015-06-06 22:57:12Z'))
        self.assertThat(
            format_value(str(now), eliot_ns(u'timestamp')),
            ExactlyEquals(u'2015-06-06 22:57:12Z'))

    def test_timestamp_field_local(self):
        """
        Format Eliot ``timestamp`` fields as human-readable local time if the
        feature was requested.
        """
        format_value = _default_value_formatter(
            human_readable=True, field_limit=0, utc_timestamps=False)
        # datetime(2015, 6, 6, 22, 57, 12)
        now = 1433631432 + time.timezone
        self.assertThat(
            format_value(now, eliot_ns(u'timestamp')),
            ExactlyEquals(u'2015-06-06 22:57:12'))
        self.assertThat(
            format_value(str(now), eliot_ns(u'timestamp')),
            ExactlyEquals(u'2015-06-06 22:57:12'))

    def test_not_eliot_timestamp_field(self):
        """
        Do not format user fields named ``timestamp``.
        """
        format_value = _default_value_formatter(
            human_readable=True, field_limit=0)
        now = 1433631432
        self.assertThat(
            format_value(now, u'timestamp'),
            ExactlyEquals(text_type(now)))
        self.assertThat(
            format_value(text_type(now), u'timestamp'),
            ExactlyEquals(text_type(now)))

    def test_timestamp_field_not_human(self):
        """
        Do not format ``timestamp`` fields as human-readable if the feature was
        not requested.
        """
        format_value = _default_value_formatter(
            human_readable=False, field_limit=0)
        # datetime(2015, 6, 6, 22, 57, 12)
        now = 1433631432
        self.assertThat(
            format_value(now, u'timestamp'),
            ExactlyEquals(text_type(now)))


colors = get_theme(dark_background=True, colored=colored)
no_colors = get_theme(dark_background=True, colored=None)


def no_formatting(value, field_name=None):
    return text_type(value)


class MessageNameTests(TestCase):
    """
    Tests for `eliottree.render.message_name`.
    """
    def test_action_type(self):
        """
        Action types include their type name.
        """
        message = next(tasks_from_iterable([action_task])).root().start_message
        self.assertThat(
            message_name(colors, no_formatting, message, options=Options()),
            StartsWith(colors.parent(message.contents.action_type)))

    def test_action_task_level(self):
        """
        Action types include their task level.
        """
        message = next(tasks_from_iterable([action_task])).root().start_message
        self.assertThat(
            message_name(colors, no_formatting, message, options=Options()),
            Contains(message.task_level.to_string()))

    def test_action_status(self):
        """
        Action types include their status.
        """
        message = next(tasks_from_iterable([action_task])).root().start_message
        self.assertThat(
            message_name(colors, no_formatting, message, options=Options()),
            Contains(u'started'))

    def test_action_status_success(self):
        """
        Successful actions color their status.
        """
        message = next(tasks_from_iterable([
            action_task, action_task_end,
        ])).root().end_message
        self.assertThat(
            message_name(colors, no_formatting, message, options=Options()),
            Contains(colors.status_success(u'succeeded')))

    def test_action_status_failed(self):
        """
        Failed actions color their status.
        """
        message = next(tasks_from_iterable([
            action_task, action_task_end_failed,
        ])).root().end_message
        self.assertThat(
            message_name(colors, no_formatting, message, options=Options()),
            Contains(colors.status_failure(u'failed')))

    def test_message_type(self):
        """
        Message types include their type name.
        """
        message = WrittenMessage.from_dict(message_task)
        self.assertThat(
            message_name(colors, no_formatting, message, options=Options()),
            StartsWith(colors.parent(message.contents.message_type)))

    def test_message_task_level(self):
        """
        Message types include their task level.
        """
        message = WrittenMessage.from_dict(message_task)
        self.assertThat(
            message_name(colors, no_formatting, message, options=Options()),
            Contains(message.task_level.to_string()))

    def test_unknown(self):
        """
        None or messages with no identifiable information are rendered as
        ``<unnamed>``.
        """
        self.assertThat(
            message_name(colors, no_formatting, None, options=Options()),
            ExactlyEquals(u'<unnamed>'))
        message = WrittenMessage.from_dict({u'timestamp': 0})
        self.assertThat(
            message_name(colors, no_formatting, message, options=Options()),
            ExactlyEquals(u'<unnamed>'))


class FormatNodeTests(TestCase):
    """
    Tests for `eliottree.render.format_node`.
    """
    def format_node(self, node, format_value=None, colors=no_colors):
        if format_value is None:
            def format_value(value, field_name=None):
                return value
        return format_node(format_value, colors, Options(), node)

    def test_task(self):
        """
        `Task`'s UUID is rendered.
        """
        node = next(tasks_from_iterable([action_task]))
        self.assertThat(
            self.format_node(node),
            ExactlyEquals(node.root().task_uuid))

    def test_written_action(self):
        """
        `WrittenAction`'s start message is rendered.
        """
        tasks = tasks_from_iterable([action_task, action_task_end])
        node = next(tasks).root()
        self.assertThat(
            self.format_node(node),
            ExactlyEquals(u'{}{} {} {} {} \u29d6 {}'.format(
                node.start_message.contents.action_type,
                node.start_message.task_level.to_string(),
                RIGHT_DOUBLE_ARROW,
                node.start_message.contents.action_status,
                node.start_message.timestamp,
                node.end_message.timestamp - node.start_message.timestamp)))

    def test_tuple_list(self):
        """
        Tuples can be a key and list, in which case the value is not rendered
        here.
        """
        node = (u'a\nb\x1bc', [u'x\n', u'y\x1b', u'z'])
        self.assertThat(
            self.format_node(node, colors=colors),
            ExactlyEquals(u'{}: {}'.format(
                colors.prop_key(u'a\u240ab\u241bc'),
                colors.prop_value(u''))))

    def test_tuple_dict(self):
        """
        Tuples can be a key and dict, in which case the value is not rendered
        here.
        """
        node = (u'a\nb\x1bc', {u'x\n': u'y\x1b', u'z': u'zz'})
        self.assertThat(
            self.format_node(node, colors=colors),
            ExactlyEquals(u'{}: {}'.format(
                colors.prop_key(u'a\u240ab\u241bc'),
                colors.prop_value(u''))))

    def test_tuple_other(self):
        """
        Tuples can be a key and string, number, etc. rendered inline.
        """
        node = (u'a\nb\x1bc', u'hello')
        self.assertThat(
            self.format_node(node, colors=colors),
            ExactlyEquals(u'{}: {}'.format(
                colors.prop_key(u'a\u240ab\u241bc'),
                colors.prop_value(u'hello'))))

        node = (u'a\nb\x1bc', 42)
        self.assertThat(
            self.format_node(node, colors=colors),
            ExactlyEquals(u'{}: {}'.format(
                colors.prop_key(u'a\u240ab\u241bc'),
                colors.prop_value(u'42'))))

    def test_other(self):
        """
        Other types raise `NotImplementedError`.
        """
        node = object()
        with ExpectedException(NotImplementedError):
            self.format_node(node, colors=colors)


class MessageFieldsTests(TestCase):
    """
    Tests for `eliottree.render.message_fields`.
    """
    def test_empty(self):
        """
        ``None`` or empty messages result in no fields.
        """
        self.assertThat(
            message_fields(None, set()),
            Equals([]))
        message = WrittenMessage.from_dict({})
        self.assertThat(
            message_fields(message, set()),
            Equals([]))

    def test_fields(self):
        """
        Include all the message fields but not the timestamp.
        """
        message = WrittenMessage.from_dict({u'a': 1})
        self.assertThat(
            message_fields(message, set()),
            Equals([(u'a', 1)]))
        message = WrittenMessage.from_dict({u'a': 1, u'timestamp': 12345678})
        self.assertThat(
            message_fields(message, set()),
            Equals([
                (u'a', 1)]))

    def test_ignored_fields(self):
        """
        Ignore any specified fields.
        """
        message = WrittenMessage.from_dict({u'a': 1, u'b': 2})
        self.assertThat(
            message_fields(message, {u'b'}),
            Equals([(u'a', 1)]))


class GetChildrenTests(TestCase):
    """
    Tests for `eliottree.render.get_children`.
    """
    def test_task_action(self):
        """
        The root message is the only child of a `Task`.
        """
        node = next(tasks_from_iterable([action_task]))
        self.assertThat(
            get_children(set(), node),
            Equals([node.root()]))
        node = next(tasks_from_iterable([message_task]))
        self.assertThat(
            get_children(set(), node),
            Equals([node.root()]))

    def test_written_action_ignored_fields(self):
        """
        Action fields can be ignored.
        """
        task = dict(action_task)
        task.update({u'c': u'3'})
        node = next(tasks_from_iterable([task])).root()
        start_message = node.start_message
        self.assertThat(
            list(get_children({u'c'}, node)),
            Equals([
                (u'action_status', start_message.contents.action_status),
                (u'action_type', start_message.contents.action_type)]))

    def test_written_action_start(self):
        """
        The children of a `WrittenAction` begin with the fields of the start
        message.
        """
        node = next(tasks_from_iterable([
            action_task, nested_action_task, action_task_end])).root()
        start_message = node.start_message
        self.assertThat(
            list(get_children({u'foo'}, node))[:2],
            Equals([
                (u'action_status', start_message.contents.action_status),
                (u'action_type', start_message.contents.action_type)]))

    def test_written_action_children(self):
        """
        The children of a `WrittenAction` contain child actions/messages.
        """
        node = next(tasks_from_iterable([
            action_task, nested_action_task, action_task_end])).root()
        self.assertThat(
            list(get_children({u'foo'}, node))[2],
            Equals(node.children[0]))

    def test_written_action_no_children(self):
        """
        The children of a `WrittenAction` does not contain any child
        actions/messages if there aren't any.
        """
        node = next(tasks_from_iterable([action_task])).root()
        self.assertThat(
            list(get_children({u'foo'}, node)),
            HasLength(2))

    def test_written_action_no_end(self):
        """
        If the `WrittenAction` has no end message, it is excluded.
        """
        node = next(tasks_from_iterable([
            action_task, nested_action_task])).root()
        self.assertThat(
            list(get_children({u'foo'}, node))[4:],
            Equals([]))

    def test_written_action_end(self):
        """
        The last child of a `WrittenAction` is the end message.
        """
        node = next(tasks_from_iterable([
            action_task, nested_action_task, action_task_end])).root()
        self.assertThat(
            list(get_children({u'foo'}, node))[3:],
            Equals([node.end_message]))

    def test_written_message(self):
        """
        The fields of `WrittenMessage`\\s are their children. Fields can also be
        ignored.
        """
        node = WrittenMessage.from_dict({u'a': 1, u'b': 2, u'c': 3})
        self.assertThat(
            get_children({u'c'}, node),
            Equals([
                (u'a', 1),
                (u'b', 2)]))

    def test_tuple_dict(self):
        """
        The key/value pairs of dicts are their children.
        """
        node = (u'key', {u'a': 1, u'b': 2, u'c': 3})
        self.assertThat(
            # The ignores intentionally do nothing.
            get_children({u'c'}, node),
            Equals([
                (u'a', 1),
                (u'b', 2),
                (u'c', 3)]))

    def test_tuple_list(self):
        """
        The indexed items of lists are their children.
        """
        node = (u'key', [u'a', u'b', u'c'])
        self.assertThat(
            # The ignores intentionally do nothing.
            get_children({2, u'c'}, node),
            After(list,
                  Equals([
                      (0, u'a'),
                      (1, u'b'),
                      (2, u'c')])))

    def test_other(self):
        """
        Other values are considered to have no children.
        """
        self.assertThat(get_children({}, None), Equals([]))
        self.assertThat(get_children({}, 42), Equals([]))
        self.assertThat(get_children({}, u'hello'), Equals([]))


class RenderTasksTests(TestCase):
    """
    Tests for `eliottree.render.render_tasks`.
    """
    def render_tasks(self, iterable, **kw):
        fd = StringIO()
        err = StringIO(u'')
        tasks = tasks_from_iterable(iterable)
        render_tasks(write=fd.write, write_err=err.write, tasks=tasks, **kw)
        if err.tell():
            return fd.getvalue(), err.getvalue()
        return fd.getvalue()

    def test_format_node_failures(self):
        """
        Catch exceptions when formatting nodes and display a message without
        interrupting the processing of tasks. List all caught exceptions to
        stderr.
        """
        def bad_format_node(*a, **kw):
            raise ValueError('Nope')
        self.assertThat(
            self.render_tasks([message_task],
                              format_node=bad_format_node),
            MatchesListwise([
                Contains(u'<node formatting exception>'),
                MatchesAll(
                    Contains(u'Traceback (most recent call last):'),
                    Contains(u'ValueError: Nope'))]))

    def test_format_value_failures(self):
        """
        Catch exceptions when formatting node values and display a message
        without interrupting the processing of tasks. List all caught
        exceptions to stderr.
        """
        def bad_format_value(*a, **kw):
            raise ValueError('Nope')
        self.assertThat(
            self.render_tasks([message_task],
                              format_value=bad_format_value),
            MatchesListwise([
                Contains(u'message: <value formatting exception>'),
                MatchesAll(
                    Contains(u'Traceback (most recent call last):'),
                    Contains(u'ValueError: Nope'))]))

    def test_tasks(self):
        """
        Render two tasks of sequential levels, by default most standard Eliot
        task keys are ignored.
        """
        self.assertThat(
            self.render_tasks([action_task, action_task_end]),
            ExactlyEquals(
                u'f3a32bb3-ea6b-457c-aa99-08a3d0491ab4\n'
                u'\u2514\u2500\u2500 app:action/1 \u21d2 started '
                u'1425356800 \u29d6 2\n'
                u'    \u2514\u2500\u2500 app:action/2 \u21d2 succeeded '
                u'1425356802\n\n'))

    def test_tasks_human_readable(self):
        """
        Render two tasks of sequential levels, by default most standard Eliot
        task keys are ignored, values are formatted to be human readable.
        """
        self.assertThat(
            self.render_tasks([action_task, action_task_end],
                              human_readable=True),
            ExactlyEquals(
                u'f3a32bb3-ea6b-457c-aa99-08a3d0491ab4\n'
                u'\u2514\u2500\u2500 app:action/1 \u21d2 started '
                u'2015-03-03 04:26:40Z \u29d6 2.000s\n'
                u'    \u2514\u2500\u2500 app:action/2 \u21d2 succeeded '
                u'2015-03-03 04:26:42Z\n'
                u'\n'))

    def test_multiline_field(self):
        """
        When no field limit is specified for task values, multiple lines are
        output for multiline tasks.
        """
        fd = StringIO()
        tasks = tasks_from_iterable([multiline_action_task])
        render_tasks(
            write=fd.write,
            tasks=tasks)
        self.assertThat(
            fd.getvalue(),
            ExactlyEquals(
                u'f3a32bb3-ea6b-457c-aa99-08a3d0491ab4\n'
                u'\u2514\u2500\u2500 app:action/1 \u21d2 started '
                u'1425356800\n'
                u'    \u2514\u2500\u2500 message: this is a\u23ce\n'
                u'        many line message\n\n'))

    def test_multiline_field_limit(self):
        """
        When a field limit is specified for task values, only the first of
        multiple lines is output.
        """
        self.assertThat(
            self.render_tasks([multiline_action_task],
                              field_limit=1000),
            ExactlyEquals(
                u'f3a32bb3-ea6b-457c-aa99-08a3d0491ab4\n'
                u'\u2514\u2500\u2500 app:action/1 \u21d2 started '
                u'1425356800\n'
                u'    \u2514\u2500\u2500 message: this is a\u2026\n\n'))

    def test_field_limit(self):
        """
        Truncate task values that are longer than the field_limit if specified.
        """
        self.assertThat(
            self.render_tasks([message_task],
                              field_limit=5),
            ExactlyEquals(
                u'cdeb220d-7605-4d5f-8341-1a170222e308\n'
                u'\u2514\u2500\u2500 twisted:log/1 '
                u'14253\u2026\n'
                u'    \u251c\u2500\u2500 error: False\n'
                u'    \u2514\u2500\u2500 message: Main \u2026\n\n'))

    def test_ignored_keys(self):
        """
        Task keys can be ignored.
        """
        self.assertThat(
            self.render_tasks([action_task],
                              ignored_fields={u'action_type'}),
            ExactlyEquals(
                u'f3a32bb3-ea6b-457c-aa99-08a3d0491ab4\n'
                u'\u2514\u2500\u2500 app:action/1 \u21d2 started '
                u'1425356800\n'
                u'    \u2514\u2500\u2500 action_status: started\n\n'))

    def test_task_data(self):
        """
        Task data is rendered as tree elements.
        """
        self.assertThat(
            self.render_tasks([message_task]),
            ExactlyEquals(
                u'cdeb220d-7605-4d5f-8341-1a170222e308\n'
                u'\u2514\u2500\u2500 twisted:log/1 '
                u'1425356700\n'
                u'    \u251c\u2500\u2500 error: False\n'
                u'    \u2514\u2500\u2500 message: Main loop terminated.\n\n'))

    def test_dict_data(self):
        """
        Task values that are ``dict``s are rendered as tree elements.
        """
        self.assertThat(
            self.render_tasks([dict_action_task]),
            ExactlyEquals(
                u'f3a32bb3-ea6b-457c-aa99-08a3d0491ab4\n'
                u'\u2514\u2500\u2500 app:action/1 \u21d2 started '
                u'1425356800\n'
                u'    \u2514\u2500\u2500 some_data: \n'
                u'        \u2514\u2500\u2500 a: 42\n\n'))

    def test_list_data(self):
        """
        Task values that are ``list``s are rendered as tree elements.
        """
        self.assertThat(
            self.render_tasks([list_action_task]),
            ExactlyEquals(
                u'f3a32bb3-ea6b-457c-aa99-08a3d0491ab4\n'
                u'\u2514\u2500\u2500 app:action/1 \u21d2 started '
                u'1425356800\n'
                u'    \u2514\u2500\u2500 some_data: \n'
                u'        \u251c\u2500\u2500 0: a\n'
                u'        \u2514\u2500\u2500 1: b\n\n'))

    def test_nested(self):
        """
        Render nested tasks in a way that visually represents that nesting.
        """
        self.assertThat(
            self.render_tasks([action_task, nested_action_task]),
            ExactlyEquals(
                u'f3a32bb3-ea6b-457c-aa99-08a3d0491ab4\n'
                u'\u2514\u2500\u2500 app:action/1 \u21d2 started '
                u'1425356800\n'
                u'    \u2514\u2500\u2500 app:action:nest/1/1 \u21d2 started '
                u'1425356900\n\n'))

    def test_janky_message(self):
        """
        Task names, UUIDs, keys and values in messages all have control
        characters escaped.
        """
        self.assertThat(
            self.render_tasks([janky_message_task]),
            ExactlyEquals(
                u'cdeb220d-7605-4d5f-\u241b(08341-1a170222e308\n'
                u'\u2514\u2500\u2500 M\u241b(0/1 '
                u'1425356700\n'
                u'    \u251c\u2500\u2500 er\u241bror: False\n'
                u'    \u2514\u2500\u2500 mes\u240asage: '
                u'Main loop\u241b(0terminated.\n\n'))

    def test_janky_action(self):
        """
        Task names, UUIDs, keys and values in actions all have control
        characters escaped.
        """
        self.assertThat(
            self.render_tasks([janky_action_task]),
            ExactlyEquals(
                u'f3a32bb3-ea6b-457c-\u241b(0aa99-08a3d0491ab4\n'
                u'\u2514\u2500\u2500 A\u241b(0/1 \u21d2 started '
                u'1425356800\u241b(0\n'
                u'    \u251c\u2500\u2500 \u241b(0: \n'
                u'    \u2502   \u2514\u2500\u2500 \u241b(0: nope\n'
                u'    \u2514\u2500\u2500 mes\u240asage: hello\u241b(0world\n\n'
            ))

    def test_colorize(self):
        """
        Passing ``theme`` will colorize the output.
        """
        self.assertThat(
            self.render_tasks([action_task, action_task_end],
                              theme=colors),
            ExactlyEquals(
                u'\n'.join([
                    colors.root(u'f3a32bb3-ea6b-457c-aa99-08a3d0491ab4'),
                    u'\u2514\u2500\u2500 {}{} \u21d2 {} {} {} {}'.format(
                        colors.parent(u'app:action'),
                        colors.task_level(u'/1'),
                        colors.status_success(u'started'),
                        colors.timestamp(u'1425356800'),
                        HOURGLASS,
                        colors.duration(u'2')),
                    u'    \u2514\u2500\u2500 {}{} \u21d2 {} {}'.format(
                        colors.parent(u'app:action'),
                        colors.task_level(u'/2'),
                        colors.status_success(u'succeeded'),
                        colors.timestamp(u'1425356802')),
                    u'\n',
                ])))



from eliottree._errors import EliotParseError, JSONParseError
from eliottree._parse import tasks_from_iterable
from eliottree._render import render_tasks
from eliottree.filter import (
    filter_by_end_date, filter_by_jmespath, filter_by_start_date,
    filter_by_uuid, combine_filters_and)
from eliottree._theme import get_theme, apply_theme_overrides, Theme
from eliottree._color import color_factory, colored


__all__ = [
    'filter_by_jmespath', 'filter_by_uuid', 'filter_by_start_date',
    'filter_by_end_date', 'render_tasks', 'tasks_from_iterable',
    'EliotParseError', 'JSONParseError', 'combine_filters_and',
    'get_theme', 'apply_theme_overrides', 'Theme', 'color_factory',
    'colored',
]

from ._version import get_versions  # noqa: E402
__version__ = get_versions()['version']
del get_versions



import sys

from eliot.parse import Parser

from eliottree._errors import EliotParseError


def tasks_from_iterable(iterable):
    """
    Parse an iterable of Eliot message dictionaries into tasks.

    :type iterable: ``Iterable[Dict]``
    :param iterable: Iterable of serialized Eliot message dictionaries.
    :rtype: ``Iterable``
    :return: Iterable of parsed Eliot tasks, suitable for use with
    `eliottree.render_tasks`.
    """
    parser = Parser()
    for message_dict in iterable:
        try:
            completed, parser = parser.add(message_dict)
            for task in completed:
                yield task
        except Exception:
            raise EliotParseError(message_dict, sys.exc_info())
    for task in parser.incomplete_tasks():
        yield task


__all__ = ['tasks_from_iterable']
